package com.letv.shop.lekafkavalve;

import java.lang.reflect.Field;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpSession;

import org.apache.catalina.AccessLog;
import org.apache.catalina.Session;
import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.coyote.RequestInfo;
import org.apache.tomcat.util.ExceptionUtils;

/**
 * @Description:
 * @Copyright: by Author Dimmacro
 * @Comany:LETV
 * @Create time: 2016年5月30日下午3:06:46
 * 
 */
public abstract class AccessLogElement {

	protected abstract String buildLog(Request request, Response response, long time, AccessLog accesslog);
	
	/**
	 * @Description:
	 * @param pattern
	 * @return
	 * @Author:denghong1 2016年5月30日下午3:23:57
	 * @update1:
	 * 
	 */
	public static AccessLogElement parsePattern(String pattern) {
		final List<AccessLogElement> list = new ArrayList<>();
		boolean replace = false;
		StringBuilder buf = new StringBuilder();
		for (int i = 0; i < pattern.length(); ++i) {
			char ch = pattern.charAt(i);
			if (replace) {
				if ('{' == ch) {
					StringBuilder name = new StringBuilder();
					int j = i + 1;
					for (; (j < pattern.length()) && ('}' != pattern.charAt(j)); ++j) {
						name.append(pattern.charAt(j));
					}
					if (j + 1 < pattern.length()) {
						++j;
						list.add(createAccessLogElement(name.toString(), pattern.charAt(j)));

						i = j;
					} else {
						list.add(createAccessLogElement(ch));
					}
				} else {
					list.add(createAccessLogElement(ch));
				}
				replace = false;
			} else if (ch == '%') {
				replace = true;
				list.add(new StringElement(buf.toString()));
				buf = new StringBuilder();
			} else {
				buf.append(ch);
			}
		}
		if (buf.length() > 0) {
			list.add(new StringElement(buf.toString()));
		}
		return new AccessLogElement() {
			@Override
			protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
				StringBuilder sBuilder = new StringBuilder(30);
				for (AccessLogElement accessLogElement : list) {
					sBuilder.append(accessLogElement.buildLog(request, response, time, accesslog));
				}
				return sBuilder.toString();
			}
		};
	}

	public static AccessLogElement createAccessLogElement(String header, char pattern) {
		switch (pattern) {
		case 'i':
			return new HeaderElement(header);
		case 'c':
			return new CookieElement(header);
		case 'o':
			return new ResponseHeaderElement(header);
		case 'r':
			return new RequestAttributeElement(header);
		case 's':
			return new SessionAttributeElement(header);
		case 't':
			return new DateAndTimeElement(header);
		case 'd':
		case 'e':
		case 'f':
		case 'g':
		case 'h':
		case 'j':
		case 'k':
		case 'l':
		case 'm':
		case 'n':
		case 'p':
		case 'q':
		}
		return new StringElement("???");
	}

	public static AccessLogElement createAccessLogElement(char pattern) {
		switch (pattern) {
		case 'a':
			return new RemoteAddrElement();
		case 'A':
			return new LocalAddrElement();
		case 'b':
			return new ByteSentElement(true);
		case 'B':
			return new ByteSentElement(false);
		case 'D':
			return new ElapsedTimeElement(true);
		case 'F':
			return new FirstByteTimeElement();
		case 'h':
			return new HostElement();
		case 'H':
			return new ProtocolElement();
		case 'l':
			return new LogicalUserNameElement();
		case 'm':
			return new MethodElement();
		case 'p':
			return new LocalPortElement();
		case 'q':
			return new QueryElement();
		case 'r':
			return new RequestElement();
		case 's':
			return new HttpStatusCodeElement();
		case 'S':
			return new SessionIdElement();
		case 't':
			return new DateAndTimeElement();
		case 'T':
			return new ElapsedTimeElement(false);
		case 'u':
			return new UserElement();
		case 'U':
			return new RequestURIElement();
		case 'v':
			return new LocalServerNameElement();
		case 'I':
			return new ThreadNameElement();
		case 'C':
		case 'E':
		case 'G':
		case 'J':
		case 'K':
		case 'L':
		case 'M':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '[':
		case '\\':
		case ']':
		case '^':
		case '_':
		case '`':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'g':
		case 'i':
		case 'j':
		case 'k':
		case 'n':
		case 'o':
		}
		return new StringElement(new StringBuilder().append("???").append(pattern).append("???").toString());
	}

	protected static class SessionAttributeElement extends AccessLogElement {
		private final String header;

		public SessionAttributeElement(String header) {
			this.header = header;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accessLog) {
			Object value = null;
			if (null != request) {
				HttpSession sess = request.getSession(false);
				if (null != sess)
					value = sess.getAttribute(this.header);
			} else {
				value = "??";
			}
			return null == value ? "-" : value.toString();
		}
	}

	protected static class RequestAttributeElement extends AccessLogElement {
		private final String header;

		public RequestAttributeElement(String header) {
			this.header = header;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			Object value = null;
			if (request != null)
				value = request.getAttribute(this.header);
			else {
				value = "??";
			}
			return null == value ? "-" : value.toString();
		}
	}

	protected static class ResponseHeaderElement extends AccessLogElement {
		private final String header;

		public ResponseHeaderElement(String header) {
			this.header = header;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (null != response) {
				Iterator iter = response.getHeaders(this.header).iterator();
				if (iter.hasNext()) {
					StringBuilder buf = new StringBuilder();
					buf.append((String) iter.next());
					while (iter.hasNext()) {
						buf.append(',').append((String) iter.next());
					}
					return buf.toString();
				}
			}
			return "-";
		}
	}

	protected static class CookieElement extends AccessLogElement {
		private final String header;

		public CookieElement(String header) {
			this.header = header;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			Cookie[] c = request.getCookies();
			if (c != null) {
				for (int i = 0; i < c.length; ++i) {
					if (this.header.equals(c[i].getName())) {
						return c[i].getValue();
					}
				}
			}
			return "-";
		}
	}

	protected static class HeaderElement extends AccessLogElement {
		private final String header;

		public HeaderElement(String header) {
			this.header = header;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			Enumeration iter = request.getHeaders(this.header);
			if (iter.hasMoreElements()) {
				StringBuilder buf = new StringBuilder();
				buf.append((String) iter.nextElement());
				while (iter.hasMoreElements()) {
					buf.append(',').append((String) iter.nextElement());
				}
				return buf.toString();
			}
			return "-";
		}
	}

	protected static class StringElement extends AccessLogElement {
		private final String str;

		public StringElement(String str) {
			this.str = str;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			return this.str;
		}
	}

	protected static class LocalServerNameElement extends AccessLogElement {
		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			return request.getServerName();
		}
	}

	protected static class RequestURIElement extends AccessLogElement {
		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (request != null) {
				return request.getRequestURI();
			} else {
				return "-";
			}
		}
	}

	protected static class SessionIdElement extends AccessLogElement {
		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (request == null) {
				return "-";
			} else {
				Session session = request.getSessionInternal(false);
				if (session == null)
					return "-";
				else
					return session.getIdInternal();
			}
		}
	}

	protected static class QueryElement extends AccessLogElement {
		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			String query = null;
			if (request != null) {
				query = request.getQueryString();
			}
			return null == query ? "" : ("?" + query);
		}
	}

	protected static class FirstByteTimeElement extends AccessLogElement {
		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			long commitTime = response.getCoyoteResponse().getCommitTime();
			if (commitTime == -1L) {
				return "-";
			} else {
				long delta = commitTime - request.getCoyoteRequest().getStartTime();
				return Long.toString(delta);
			}
		}

	}

	protected static class ElapsedTimeElement extends AccessLogElement {
		private final boolean millis;

		public ElapsedTimeElement(boolean millis) {
			this.millis = millis;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (this.millis) {
				return "" + time;
			} else {
				StringBuilder buf = new StringBuilder();
				buf.append(time / 1000L);
				buf.append('.');
				int remains = (int) (time % 1000L);
				buf.append(remains / 100);
				remains %= 100;
				buf.append(remains / 10);
				buf.append(remains % 10);
				return buf.toString();
			}
		}
	}

	protected static class MethodElement extends AccessLogElement {
		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			return null == request ? "" : request.getMethod();
		}
	}

	protected static class ByteSentElement extends AccessLogElement {
		private final boolean conversion;

		public ByteSentElement(boolean conversion) {
			this.conversion = conversion;
		}

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			long length = response.getBytesWritten(false);
			if (length <= 0L) {
				Object start = request.getAttribute("org.apache.tomcat.sendfile.start");

				if (start instanceof Long) {
					Object end = request.getAttribute("org.apache.tomcat.sendfile.end");

					if (end instanceof Long) {
						length = ((Long) end).longValue() - ((Long) start).longValue();
					}
				}
			}

			if ((length <= 0L) && (this.conversion))
				return "-";
			else
				return "" + length;
		}
	}

	protected static class LocalPortElement extends AccessLogElement {

		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (accesslog.getRequestAttributesEnabled()) {
				Object port = request.getAttribute("org.apache.catalina.AccessLog.ServerPort");
				if (port == null)
					return "" + request.getServerPort();
				else
					return "" + port;
			} else {
				return "" + request.getServerPort();
			}
		}
	}

	protected static class HttpStatusCodeElement extends AccessLogElement {
		@Override
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			return null == response ? "-": "" + response.getStatus() ;
		}
	}

	protected static class RequestElement extends AccessLogElement {
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (request != null) {
				String method = request.getMethod();
				if (method == null) {
					return "-";
				} else {
					StringBuilder buf = new StringBuilder();
					buf.append(request.getMethod());
					buf.append(' ');
					buf.append(request.getRequestURI());
					if (request.getQueryString() != null) {
						buf.append('?');
						buf.append(request.getQueryString());
					}
					buf.append(' ');
					buf.append(request.getProtocol());
					return buf.toString();
				}
			} else {
				return "-";
			}
		}
	}

	protected static enum FormatType {
		CLF, SEC, MSEC, MSEC_FRAC, SDF;
	}

	protected static class DateAndTimeElement extends AccessLogElement {
		private static final String requestStartPrefix = "begin";
		private static final String responseEndPrefix = "end";
		private static final String prefixSeparator = ":";
		private static final String secFormat = "sec";
		private static final String msecFormat = "msec";
		private static final String msecFractionFormat = "msec_frac";
		private static final String msecPattern = "{#}";
		private static final String trippleMsecPattern = "{#}{#}{#}";
		private String format;
		private boolean usesBegin;
		private AccessLogElement.FormatType type;
		private boolean usesMsecs;

		// 存储accesslog的locale，只存一次
		private Locale accessLogLocale;

		public DateAndTimeElement() {
			this(null);
		}

		private String tidyFormat(String format) {
			boolean escape = false;
			StringBuilder result = new StringBuilder();
			int len = format.length();

			for (int i = 0; i < len; ++i) {
				char x = format.charAt(i);
				if ((escape) || (x != 'S')) {
					result.append(x);
				} else {
					result.append("{#}");
					this.usesMsecs = true;
				}
				if (x == '\'') {
					escape = !(escape);
				}
			}
			return result.toString();
		}

		public DateAndTimeElement(String paramString) {
			this.format = null;
			this.usesBegin = false;
			this.type = FormatType.CLF;
			this.usesMsecs = false;
			this.format = paramString;
			if (this.format != null) {
				if (this.format.equals(requestStartPrefix)) {
					this.usesBegin = true;
					this.format = "";
				} else if (this.format.startsWith(requestStartPrefix+prefixSeparator)) {
					this.usesBegin = true;
					this.format = this.format.substring(6);
				} else if (this.format.equals(responseEndPrefix)) {
					this.usesBegin = false;
					this.format = "";
				} else if (this.format.startsWith(responseEndPrefix+prefixSeparator)) {
					this.usesBegin = false;
					this.format = this.format.substring(4);
				}
				if (this.format.length() == 0) {
					this.type = FormatType.CLF;
				} else if (this.format.equals(secFormat)) {
					this.type = FormatType.SEC;
				} else if (this.format.equals(msecFormat)) {
					this.type = FormatType.MSEC;
				} else if (this.format.equals(msecFractionFormat)) {
					this.type = FormatType.MSEC_FRAC;
				} else {
					this.type = FormatType.SDF;
					this.format = tidyFormat(this.format);
				}
			}
		}

		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			long timestamp = request.getCoyoteRequest().getStartTime() + time;
			StringBuilder buf = new StringBuilder();
			if (this.usesBegin)
				timestamp -= time;
			long frac;
			switch (this.type.ordinal()) {
			case 0:
				buf.append(DateFormatCache.getFormatDate(timestamp));
				break;
			case 1:
				buf.append(timestamp / 1000L);
				break;
			case 2:
				buf.append(timestamp);
				break;
			case 3:
				frac = timestamp % 1000L;
				if (frac < 100L) {
					if (frac < 10L) {
						buf.append('0');
						buf.append('0');
					} else {
						buf.append('0');
					}
				}
				buf.append(frac);
				break;
			case 4:
				if (null == this.accessLogLocale) {
					this.accessLogLocale = resolveLocale(accesslog);
				}
				String temp = DateFormatCache.localDateCache.get().getFormat(this.format, this.accessLogLocale, timestamp);

				if (this.usesMsecs) {
					frac = timestamp % 1000L;
					StringBuilder trippleMsec = new StringBuilder(4);
					if (frac < 100L) {
						if (frac < 10L) {
							trippleMsec.append('0');
							trippleMsec.append('0');
						} else {
							trippleMsec.append('0');
						}
					}
					trippleMsec.append(frac);
					temp = temp.replace(trippleMsecPattern, trippleMsec);
					temp = temp.replace(msecPattern, Long.toString(frac));
				}
				buf.append(temp);
			}
			return buf.toString();
		}

		/**
		 * @Description:
		 * @param accesslog
		 * @return
		 * @Author:denghong1 2016年5月31日上午10:50:25
		 * @update1:
		 * 
		 */
		private Locale resolveLocale(AccessLog accesslog) {
			try {
				Class<? extends AccessLog> accessLogCalss = accesslog.getClass();
				Field localeField = accessLogCalss.getDeclaredField("locale");
				localeField.setAccessible(true);
				Class<?> fieldClass = localeField.getType();
				if (Locale.class.equals(fieldClass)) {
					return Locale.class.cast(localeField.get(accesslog));
				}
			} catch (Exception e) {
				// do nothing
			}
			return Locale.getDefault();
		}

	}

	protected static class UserElement extends AccessLogElement {
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (request != null) {
				String value = request.getRemoteUser();
				if (value != null) {
					return value;
				}
			}
			return "-";
		}
	}

	protected static class ProtocolElement extends AccessLogElement {
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (accesslog.getRequestAttributesEnabled()) {
				Object proto = request.getAttribute("org.apache.catalina.AccessLog.Protocol");
				if (proto != null) {
					return proto.toString();
				}
			}
			return request.getProtocol();
		}
	}

	protected static class LogicalUserNameElement extends AccessLogElement {
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			return "-";
		}
	}

	protected static class HostElement extends AccessLogElement {
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			String value = null;
			if (accesslog.getRequestAttributesEnabled()) {
				Object host = request.getAttribute("org.apache.catalina.AccessLog.RemoteHost");
				if (host != null) {
					value = host.toString();
				}
			}
			if ((value == null) || (value.length() == 0)) {
				value = request.getRemoteHost();
			}
			if ((value == null) || (value.length() == 0)) {
				value = "-";
			}
			return value;
		}
	}

	protected static class RemoteAddrElement extends AccessLogElement {
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			if (accesslog.getRequestAttributesEnabled()) {
				Object addr = request.getAttribute("org.apache.catalina.AccessLog.RemoteAddr");
				if (addr != null) {
					return addr.toString();
				}
			}
			return request.getRemoteAddr();
		}
	}

	protected static class LocalAddrElement extends AccessLogElement {
		private static final String LOCAL_ADDR_VALUE;

		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			return LOCAL_ADDR_VALUE;
		}

		static {
			String init;
			try {
				init = getLocalIP();
			} catch (Throwable e) {
				ExceptionUtils.handleThrowable(e);
				init = "127.0.0.1";
			}
			LOCAL_ADDR_VALUE = init;
		}

		private static String getLocalIP() throws Exception {
			for (Enumeration<NetworkInterface> en = NetworkInterface.getNetworkInterfaces(); en.hasMoreElements();) {
				NetworkInterface intf = en.nextElement();
				for (Enumeration<InetAddress> enumIpAddr = intf.getInetAddresses(); enumIpAddr.hasMoreElements();) {
					InetAddress inetAddress = enumIpAddr.nextElement();
					if (!inetAddress.isLoopbackAddress() && !inetAddress.isLinkLocalAddress() && inetAddress.isSiteLocalAddress()) {
						return inetAddress.getHostAddress();
					}

				}
			}
			return "127.0.0.1";
		}
	}

	protected static class ThreadNameElement extends AccessLogElement {
		protected String buildLog(Request request, Response response, long time, AccessLog accesslog) {
			RequestInfo info = request.getCoyoteRequest().getRequestProcessor();
			if (info != null)
				return info.getWorkerThreadName();
			else
				return "-";
		}
	}
}
