package com.letv.shop.lekafkavalve;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicInteger;

import javax.servlet.ServletException;

import org.apache.catalina.AccessLog;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.LifecycleState;
import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.startup.Catalina;
import org.apache.catalina.valves.ValveBase;
import org.apache.juli.logging.Log;
import org.apache.juli.logging.LogFactory;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;

/**
 * 
 * @Description:
 * @Copyright: by Author Dimmacro
 * @Comany:LETV
 * @Create time: 2016年5月27日下午4:54:17
 * tomcat accesslog 进kafka
 */
public class LeKafkaAccesslogValve extends ValveBase implements AccessLog  {
	public static final String DEFAULT_TIMEOUT_MILLIS = "30000";
	
	private static final Log log = LogFactory.getLog(Catalina.class);
	
	protected static final String info = "com.letv.shop.lekafkavalve.LeKafkaAccesslogValve/denghong1-1.0";
	private static final String PROPERTYSPLITTER_STRING = "\\|\\|";
	private static final String KEYVALUESPLITTER_STRING = "\\=";
	
	protected boolean requestAttributesEnabled = true;
	
	private String topic;

	private String bootstrapServers;
	
	//  If set to zero then the producer will not wait for any acknowledgment from the server at all. 
	private String acks; 
	
	private String producerSize ;
	
	private String properties;
	
	private List<Producer<byte[], byte[]>> producerList;
	private AtomicInteger producerIndex = new AtomicInteger(0);

	private int timeoutMillis;

	private boolean enabled = true; // 默认配置问true，即打入kafka，除非有异常情况或主动设置了。
	
	private String pattern;

	private AccessLogElement accessLogElement;

	private String localeName;

	private Locale locale = Locale.getDefault();


	public String getPattern() {
		return pattern;
	}

	public void setPattern(String pattern) {
		if (pattern == null)
			this.pattern = "";
		else if (pattern.equals("common"))
			this.pattern = "%h %l %u %t \"%r\" %s %b";
		else if (pattern.equals("combined"))
			this.pattern = "%h %l %u %t \"%r\" %s %b \"%{Referer}i\" \"%{User-Agent}i\"";
		else {
			this.pattern = pattern;
		}
		this.accessLogElement = AccessLogElement.parsePattern(this.pattern);
	}

	/**
	 * @Description:
	 * @param paramRequest
	 * @param paramResponse
	 * @param paramLong
	 * @Author:denghong1 2016年5月27日下午4:55:39
	 * @see org.apache.catalina.AccessLog#log(org.apache.catalina.connector.Request, org.apache.catalina.connector.Response, long)
	 * @update1:
	 * 
	 */
	@Override
	public void log(Request request, Response response, long time) {
		if (producerList != null && getEnabled() && getState().isAvailable() && null != this.accessLogElement) {
            try {
				getNextProducer().send(new ProducerRecord<byte[], byte[]>(topic, this.accessLogElement.buildLog(request,response,time,this).getBytes(StandardCharsets.UTF_8))).get(timeoutMillis, TimeUnit.MILLISECONDS);
			} catch (InterruptedException | ExecutionException | TimeoutException e) {
				log.error("accesslog in kafka exception", e);
			}
        }
	}
	
	private Producer<byte[], byte[]> getNextProducer() {
    	int index = producerIndex.getAndIncrement();
    	if(index >= producerList.size()){
    		producerIndex.set(0);
    		index =0;
    	}
		return producerList.get(index);
	}

	@Override
	public void invoke(Request request, Response response) throws IOException, ServletException {
		getNext().invoke(request, response);
	}
	
	protected synchronized void startInternal() throws LifecycleException {
		initialKafkaProducer();
		setState(LifecycleState.STARTING);
	}

	/**
	 * @throws LifecycleException 
	 * @Description:
	 * @Author:denghong1 2016年5月27日下午7:07:45
	 * @update1:
	 * 
	 */
	private void initialKafkaProducer() throws LifecycleException {
		if(isEmpty(this.topic)){
			throw new LifecycleException("topic must be set");
		}
		if(isEmpty(this.bootstrapServers)){
			throw new LifecycleException("bootstrapServers must be set");
		}
		Properties config = new Properties();
		config.setProperty("key.serializer", "org.apache.kafka.common.serialization.ByteArraySerializer");
	    config.setProperty("value.serializer", "org.apache.kafka.common.serialization.ByteArraySerializer");
	    config.setProperty("batch.size", "0");
	    if(!isEmpty(this.properties)){
	    	resolveProperties(this.properties,config);
	    }
	    // 以参数设置的为准
	    config.setProperty("bootstrap.servers", this.bootstrapServers);
	    config.setProperty("acks", isEmpty(this.acks)?config.getProperty("acks", "0"):this.acks); // 明确配置的优先级高于properties配置的
	    this.timeoutMillis = Integer.parseInt(config.getProperty("timeout.ms", DEFAULT_TIMEOUT_MILLIS));
	    
	    int size = Integer.parseInt(isEmpty(this.producerSize)?config.getProperty("producer.size", "1"):this.producerSize); //确配置的优先级高于properties配置的
	    producerList = new ArrayList<>(size);
	    for(int i=0;i<size;i++){
    		producerList.add(buildKafkaProducer(config));
    	}
	}

	/**
	 * @Description:
	 * @return
	 * @Author:denghong1 2016年5月30日上午10:57:16
	 * @update1:
	 * 
	 */
	private Producer<byte[], byte[]> buildKafkaProducer(Properties config) {
		 return new KafkaProducer<>(config);
	}

	/**
	 * @Description:
	 * @param config
	 * @Author:denghong1 2016年5月27日下午8:17:12
	 * @update1:
	 * 
	 */
	private void resolveProperties(String proStr, Properties config) {
		String[] keyvaluePairs = proStr.split(PROPERTYSPLITTER_STRING);
		for (String keyvalueStr : keyvaluePairs) {
			String[] keyvaluePair = keyvalueStr.split(KEYVALUESPLITTER_STRING);
			if (keyvaluePair.length == 2) {
				config.setProperty(keyvaluePair[0], keyvaluePair[1]);
			}
		}
	}

	protected synchronized void stopInternal() throws LifecycleException {
		setState(LifecycleState.STOPPING);
		close();
	}
	

	/**
	 * @Description:
	 * @Author:denghong1 2016年5月27日下午7:04:01
	 * @update1:
	 * 
	 */
	private void close() {
		if (producerList != null) {
            // This thread is a workaround for this Kafka issue: https://issues.apache.org/jira/browse/KAFKA-1660
            final Thread closeThread = new Thread(new Runnable() {
                @Override
                public void run() {
                	for(Producer<?, ?> producer : producerList){
                		producer.close();
                	}
                }
            });
            closeThread.setName("LeKafkaAccesslogValve-CloseThread");
            closeThread.setDaemon(true); // avoid blocking JVM shutdown
            closeThread.start();
            try {
                closeThread.join(timeoutMillis);
            } catch (final InterruptedException ignore) {
                // ignore
            }
        }
		
	}

	public String getTopic() {
		return topic;
	}

	public void setTopic(String topic) {
		this.topic = topic;
	}

	public String getBootstrapServers() {
		return bootstrapServers;
	}

	public void setBootstrapServers(String bootstrapServers) {
		this.bootstrapServers = bootstrapServers;
	}

	public String getAcks() {
		return acks;
	}

	public void setAcks(String acks) {
		this.acks = acks;
	}

	public String getProducerSize() {
		return producerSize;
	}

	public void setProducerSize(String producerSize) {
		this.producerSize = producerSize;
	}

	public String getProperties() {
		return properties;
	}

	public void setProperties(String properties) {
		this.properties = properties;
	}

	public int getTimeoutMillis() {
		return timeoutMillis;
	}

	public void setTimeoutMillis(int timeoutMillis) {
		this.timeoutMillis = timeoutMillis;
	}

	public void setRequestAttributesEnabled(boolean requestAttributesEnabled) {
		this.requestAttributesEnabled = requestAttributesEnabled;
	}

	public boolean getRequestAttributesEnabled() {
		return this.requestAttributesEnabled;
	}
	
	private boolean isEmpty(String str){
		return null == str || "".equals(str.trim());
	}
	
	public boolean getEnabled() {
		return this.enabled ;
	}
	
	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}
	
	public String getLocale() {
		return this.localeName;
	}

	public void setLocale(String localeName) {
		this.localeName = localeName;
		this.locale  = findLocale(localeName, this.locale);
	}
	
	protected static Locale findLocale(String name, Locale fallback) {
		if ((name == null) || (name.isEmpty())) {
			return Locale.getDefault();
		}
		for (Locale l : Locale.getAvailableLocales()) {
			if (name.equalsIgnoreCase(l.toString())) {
				return l;
			}
		}
		log.error("LeKafkaAccesslogValve.invalidLocale:"+name);
		return fallback;
	}
	
}
